@@ -2,7 +2,7 @@ package DBIx::XHTML_Table;
 
 use strict;
 use vars qw($VERSION);
-$VERSION = '1.22';
+$VERSION = '1.36';
 
 use DBI;
 use Carp;
@@ -30,13 +30,14 @@ sub new {
 	# last arg might be GTCH (global table config hash)
 	$self->{'global'} = pop if ref $_[$#_] eq 'HASH';
 
-	# disconnected handles aren't caught :(
-	if (ref $_[0] eq 'DBI::db') {
+	# note: disconnected handles aren't caught :(
+
+	if (UNIVERSAL::isa($_[0],'DBI::db')) {
 		# use supplied db handle
 		$self->{'dbh'}        = $_[0];
 		$self->{'keep_alive'} = 1;
 	} 
-	elsif (ref $_[0] eq 'ARRAY') {
+	elsif (ref($_[0]) eq 'ARRAY') {
 		# go ahead and accept a pre-built 2d array ref
 		$self->_do_black_magic(@_);
 	}
@@ -44,9 +45,6 @@ sub new {
 		# create my own db handle
 		eval { $self->{'dbh'} = DBI->connect(@_) };
 		carp $@ and return undef if $@;
-
-		# to RaiseError or not ... should let user decided
-		#eval { $self->{'dbh'} = DBI->connect(@_,{RaiseError=>1}) };
 	}
 
 	return $self;
@@ -58,8 +56,12 @@ sub exec_query {
 	my ($self,$sql,$vars) = @_;
 
 	carp "can't call exec_query(): do database handle" unless $self->{'dbh'};
+
 	eval {
-		$self->{'sth'} = $self->{'dbh'}->prepare($sql);
+		$self->{'sth'} = (UNIVERSAL::isa($sql,'DBI::st'))
+			? $sql
+			: $self->{'dbh'}->prepare($sql)
+		;
 		$self->{'sth'}->execute(@$vars);
 	};
 	carp $@ and return undef if $@;
@@ -307,6 +309,41 @@ sub reset {
 	my ($self) = @_;
 }
 
+sub add_cols {
+	my ($self,$config) = @_;
+	$config = [$config] unless ref $config eq 'ARRAY';
+
+	foreach (@$config) {
+		next unless ref $_ eq 'HASH';
+		my ($name,$data,$pos) = @$_{(qw(name data before))};
+		my $max_pos = $self->get_col_count();
+
+		$pos  = $self->_lookup_index(ucfirst $pos || '') || $max_pos unless defined $pos && $pos =~ /^\d+$/;
+		$pos  = $max_pos if $pos > $max_pos;
+		$data = [$data] unless ref $data eq 'ARRAY';
+
+		splice(@{$self->{'fields_arry'}},$pos,0,$name);
+		$self->_reset_fields_hash();
+		splice(@$_,$pos,0,_rotate($data)) for (@{$self->{rows}});
+	}
+
+	return $self;
+}
+
+sub drop_cols {
+	my ($self,$cols) = @_;
+	$cols = $self->_refinate($cols);
+
+	foreach my $col (@$cols) {
+		my $index = delete $self->{'fields_hash'}->{$col};
+		splice(@{$self->{'fields_arry'}},$index,1);
+		$self->_reset_fields_hash();
+		splice(@$_,$index,1) for (@{$self->{'rows'}});
+	}
+
+	return $self;
+}
+
 ###################### DEPRECATED ##################################
 
 sub get_table { 
@@ -324,43 +361,6 @@ sub map_col {
 	map_cell(@_);
 }
 
-################ ABOMINATIONS TO THE DBIx NAMESPACE ################
-
-# don't use these - they belong in /dev/null
-
-# $config = [ { name => '', data => [], before => '' }, { ... }, ... ];
-sub add_cols {
-	my ($self,$config) = @_;
-	$config = [$config] unless ref $config eq 'ARRAY';
-
-	foreach (@$config) {
-		next unless ref $_ eq 'HASH';
-		my ($name,$data,$pos) = @$_{(qw(name data before))};
-		my $max_pos = $self->get_col_count();
-
-		$pos   = $self->_lookup_index($pos || '') unless $pos && $pos =~ /^\d+$/;
-		$pos ||= $max_pos;
-		$pos   = $max_pos if $pos > $max_pos;
-		$data  = [$data] unless ref $data eq 'ARRAY';
-
-		splice(@{$self->{'fields_arry'}},$pos,0,$name);
-		$self->_reset_fields_hash();
-		splice(@$_,$pos,0,_rotate($data)) for (@{$self->{rows}});
-	}
-}
-
-sub drop_cols {
-	my ($self,$cols) = @_;
-	$cols = $self->_refinate($cols);
-
-	foreach my $col (@$cols) {
-		my $index = delete $self->{'fields_hash'}->{$col};
-		splice(@{$self->{'fields_arry'}},$index,1);
-		$self->_reset_fields_hash();
-		splice(@$_,$index,1) for (@{$self->{'rows'}});
-	}
-}
-
 #################### UNDER THE HOOD ################################
 
 # repeat: it only looks complicated
@@ -369,9 +369,13 @@ sub _build_table {
 	my ($self)  = @_;
 	my $attribs = $self->{'global'}->{'table'};
 
-	my $cdata   = $self->_build_head;
-	$cdata     .= $self->_build_body   if $self->{'rows'};
-	$cdata     .= $self->_build_foot   if $self->{'totals'};
+	my ($head,$body,$foot);
+	$head = $self->_build_head;
+	$body = $self->{'rows'}   ?  $self->_build_body : '';
+	$foot = $self->{'totals'} ?  $self->_build_foot : '';
+
+	# w3c says tfoot comes before tbody ...
+	my $cdata = $head . $foot . $body;
 
 	return _tag_it('table', $attribs, $cdata) . $N;
 }
@@ -626,7 +630,7 @@ sub _tag_it {
 		$v = _rotate($v) if (ref $v eq 'ARRAY');
 		$text .= qq| \L$k\E="$v"| unless $v =~ /^$/;
 	}
-	$text .= (defined $cdata) ? ">$cdata</$name>" : '/>';
+	$text .= (defined $cdata) ? ">$cdata</\L$name\E>" : '/>';
 }
 
 # used by map_cell() and map_head()
@@ -720,7 +724,7 @@ __END__
 
 =head1 NAME
 
-DBIx::XHTML_Table - SQL query result set to XHTML table.
+DBIx::XHTML_Table - SQL query result set to XML-based HTML table.
 
 =head1 SYNOPSIS
 
@@ -732,9 +736,9 @@ DBIx::XHTML_Table - SQL query result set to XHTML table.
   my $table = DBIx::XHTML_Table->new($data_source,$usr,$pass);
 
   $table->exec_query("
-	select foo from bar
-	where baz='qux'
-	order by foo
+      select foo from bar
+      where baz='qux'
+      order by foo
   ");
 
   print $table->output();
@@ -765,8 +769,8 @@ The DBIx::XHTML_Table homepage is now available, but still under
 construction. A partially complete FAQ and Cookbook are available
 there, as well as the Tutorial, Download and Support info: 
 
-  http://unlocalhost.com/XHTML_Table/
-  http://jeffa.perlmonk.com/XHTML_Table/
+  http://www.unlocalhost.com/XHTML_Table/
+  http://jeffa.perlmonk.org/XHTML_Table/
 
 =head1 CONSTRUCTOR
 
@@ -827,7 +831,7 @@ and destroyed 'behind the scenes'. If you need to keep the database
 connection open after the XHTML_Table object is destroyed, then
 create one yourself and pass it to the constructor:
 
-  my $dbh   = DBI->connect(
+  my $dbh = DBI->connect(
     $data_source,$usr,$passwd,
     {RaiseError => 1},
   );
@@ -836,6 +840,12 @@ create one yourself and pass it to the constructor:
     # do stuff
   $dbh->disconnect;
 
+You can also use any class that isa() DBI::db object, such
+as Apache::DBI or DBIx::Password objects:
+
+  my $dbh   = DBIx::Password->connect($user);
+  my $table = DBIx::XHTML_Table->new($dbh);
+
 =item B<style_3>
 
   $obj_ref = new DBIx::XHTML_Table($rows[,$headers])
@@ -866,10 +876,10 @@ point of this module, there is nothing stopping you from
 simply hard coding it:
 
   my $rows = [
-  	[ qw(Head1 Head2 Head3) ],
-	[ qw(foo bar baz)       ],
-	[ qw(one two three)     ],
-	[ qw(un deux trois)     ]
+     [ qw(Head1 Head2 Head3) ],
+     [ qw(foo bar baz)       ],
+     [ qw(one two three)     ],
+     [ qw(un deux trois)     ]
   ];
 
   my $table = DBIx::XHTML_Table->new($rows);
@@ -897,6 +907,16 @@ bind variable or an array reference for multiple bind vars:
 	  and   baz = ?
   ',[$foo,$bar]);
 
+exec_query() also accepts a prepared DBI::st handle:
+
+  my $sth = $dbh->prepare('
+      select bar,baz from foo
+	  where bar = ?
+	  and   baz = ?
+  ');
+
+  $table->exec_query($sth,[$foo,$bar]);
+
 Consult the DBI documentation for more details on bind vars.
 
 After the query successfully executes, the results will be
@@ -1162,7 +1182,7 @@ color the cdata inside a <td> tag pair. For example:
 
   # when CSS styles will work
   $table->modify(td => {
-	 style => 'color: red',
+    style => 'color: red',
   }, [qw(first_name last_name)]);
 
 Note that the get_current_row() and get_current_col()
@@ -1223,7 +1243,7 @@ such as 'bgcolor' or 'width'.
 This method is just a more convenient way to do the
 same thing with the modify() modify.
 
-See http://unlocalhost.com/XHTML_Table/cookbook.html#5
+See http://www.unlocalhost.com/XHTML_Table/cookbook.html#5
 for more information on coloring the table.
 
 =item B<set_col_colors>
@@ -1250,7 +1270,7 @@ instead of the CSS style 'color'. For example, you
 could use 'class' or even deprecated HTML attributes
 such as 'bgcolor' or 'width'.
 
-See http://unlocalhost.com/XHTML_Table/cookbook.html#5
+See http://www.unlocalhost.com/XHTML_Table/cookbook.html#5
 for more information on coloring the table.
 
 =item B<set_null_value>
@@ -1318,8 +1338,8 @@ and supply the following:
 
   # well, and you are viewing in IE...
   $table->modify(table => {
-	  cellspacing => 0,
-	  rules       => 'groups',
+    cellspacing => 0,
+    rules       => 'groups',
   });
 
 then horizontal lines will only appear at the point where the 'grouped' 
@@ -1328,7 +1348,7 @@ inside of <table>'s. Much nicer! Add this for a nice coloring trick:
 
   # this works with or without setting a group, by the way
   $table->modify(tbody => {
-	  bgcolor => [qw(insert rotating colors here)],
+    bgcolor => [qw(insert rotating colors here)],
   });
 
 =item B<calc_totals>
@@ -1383,6 +1403,45 @@ This method is only meaningful inside a map_cell() callback; if you access
 it otherwise, you will either receive undef or the the name of the last
 column specified in your SQL statement.
 
+=item B<add_cols>
+
+   $table->add_cols(
+      { header => '', data => [], before => '' }, { ... }, ... 
+   );
+
+Going against the philosophy of only select what you need from the database,
+this sub allows you to remove whole columns. 'header' is the name of the new
+column, you will have to ucfirst yourself. It is up to you to ensure that
+that the size of 'data' is the same as the number of rows in the original
+data set. 'before' can be an index or the name of the column. For example,
+to add a new column to the beginning:
+
+   $table->add_cols({name=>'New', data=>\@rows, before => 0});
+
+add a new column to the end:
+
+   $table->add_cols({name=>'New', data=>\@rows});
+
+or somewhere in the middle:
+
+   $table->add_cols({name=>'New', data=>\@rows}, before => 'age'});
+
+or combine all three into one call:
+
+   $table->add_cols(
+      {name=>'Foo', data=>\@rows, before => 0},
+      {name=>'Bar', data=>\@rows},
+      {name=>'Baz', data=>\@rows}, before => 'Bar'},
+   );
+
+=item B<drop_cols>
+
+   $table->drop_cols([qw(foo bar 5)];
+
+Like add_cols, drop_cols goes against said 'philosophy', but it is here for
+the sake of TIMTWOTDI. Simply pass it an array ref that contains either the
+name or positions of the columns you want to drop.
+
 =back
 
 =head1 TAG REFERENCE
@@ -1406,8 +1465,18 @@ column specified in your SQL statement.
 
 =head1 BUGS
 
-If you have found a bug, please visit Best Practical
-Solution's CPAN bug tracker at http://rt.cpan.org.
+If you have found a bug, typo, etc. please visit Best Practical Solution's
+CPAN bug tracker at http://rt.cpan.org:
+
+E<lt>http://rt.cpan.org/NoAuth/Bugs.html?Dist=DBIx-XHTML_TableE<gt>
+
+or send mail to E<lt>bug-DBIx-XHTML_Table#rt.cpan.orgE<gt>
+
+(you got this far ... you can figure out how to make that
+a valid address ... and note that i won't respond to bugs
+sent to my personal address any longer)
+
+=head1 ISSUES
 
 =over 4
 
@@ -1441,6 +1510,10 @@ Stephen Nelson for documentation/code corrections.
 
 Matt Sergeant for DBIx::XML_RDB.
 
+Aaron [trs80] Johnson for convincing me into writing add and drop cols.
+
+Richard Piacentini and Tim Alexander for recommending DBIx::Password and Apache::DBI compatability and Slaven Rezic for recommending using UNIVERSAL::isa().
+
 Perl Monks for the education.
 
 =head1 SEE ALSO 
@@ -1449,29 +1522,16 @@ DBI
 
 =head1 AUTHOR 
 
-Jeffrey Hayes Anderson <captvanhalen@yahoo.com>
+Jeffrey Hayes Anderson
 
 =head1 COPYRIGHT
 
-Copyright (c) 2003 Jeffrey Hayes Anderson.
+Copyright (c) 2004 Jeffrey Hayes Anderson.
 
-Permission is hereby granted, free of charge, to any person
-obtaining a copy of this software and associated documentation
-files (the "Software"), to deal in the Software without restriction,
-including without limitation the rights to use, copy, modify,
-merge, publish, distribute, sublicense, and/or sell copies of the
-Software, and to permit persons to whom the Software is furnished to
-do so, subject to the following conditions:
+Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 
-The above copyright notice and this permission notice shall be included
-in all copies or substantial portions of the Software.
+The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 
-THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
-OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
-THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
-OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
-ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
-OTHER DEALINGS IN THE SOFTWARE.
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 =cut
